#ifndef GATCP_H
#define GATCP_H

/**
 * @file gatcp.h
 */

#include <stdint.h>

#ifndef WIN32
#include <sys/socket.h>
#endif

#include "gatime.h"

#ifdef	__cplusplus
extern "C" {
#endif

#if defined(__SunOS_5_5_1) || defined(__SunOS_5_6) || defined(WIN32)
  typedef int GA_SOCKLEN_T;
#else
  typedef socklen_t GA_SOCKLEN_T;
#endif

#define GATCP_SOCK_NAME_LEN 200

/**
 * @brief gatcp ip connection
 */
typedef struct {
    int port;//!< @brief port
	char hostip[200];//!< @brief hostip
} GATCP_IP_CONNECTION;

#define CONN_ERR_TO "TIMEOUT" //!< @brief conncection error timeout
#define CONN_ERR_COMM "COMMERR" //!< @brief connection error commerr

/**
 * @brief GATCP_CONN_INFO
 * @note Values that pertain to server only OR client only indicated
 *       via S: or C: in comment.
 */
typedef struct {
	int sockfd;	//!< @brief  Socket for all client/server communication 
	int is_server;//!< @brief  Server sets TRUE for tcp_init_conn() 
	int listenfd; //!< @brief  S: listen socket for incoming connections 
	int make_blocking; //!< @brief make_blocking
	int rcvbuf_size; //!< @brief rcvbuf_size
	int sndbuf_size; //!< @brief sndbuf_size
	int autoreconnect; //!< @brief  C: Try to reconnect on failure to send 
	int retryLimit;	//!< @brief  C: How many times to try to reconnect 
	int retryTime; //!< @brief  C: time in milli-seconds between retries 
	int retryCount; //!< @brief Accounting: DO NOT FILL THIS IN
	int isUDS; //!< @brief  Indicates socket type is UDS 
	union {
		char sock_name[GATCP_SOCK_NAME_LEN + 1];	//!< @brief Used for UDS 
		GATCP_IP_CONNECTION ip;//!< @brief ip
	} u;//!< @brief union
	char msg[100];	//!< @brief	 Used for plog WARN on reconnect failure 
	char error[196];	//!< @brief  Filled in on tcp_init_conn() failure 
	char useNL_INFO;	//!< @brief  C: Set TRUE to switch WARN to INFO on some plogs 
	char future1;		//!< @brief  Not used yet.
	char future2;		//!< @brief  Not used yet.
	char future3;		//!< @brief  Not used yet. 
} GATCP_CONN_INFO;

/**
 * @brief GATCP_MSG_HDR
 */
typedef struct {
	int version; //!< @brief version
	uint32_t/*long*/ checksum; //!< @brief  Optional field for data integrity. 
	GATIME req_time;		//!< @brief  Optional for tracking timing statistics. 
	int error;				//!< @brief  Set TRUE if data indicates error. 
	uint32_t /*unsigned long*/ dsize; //!< @brief Size of data following this structure. 
	unsigned short acctno;	//!< @brief  Optional field for quip data. 
	unsigned short receiver;//!< @brief  Optional field for quip data. 
	unsigned short port;	//!< @brief  Optional field for quip data. 
	int type;				//!< @brief  Optional field, used to identify payload. 
	int item1;				//!< @brief  Optional for application use. 
	int item2;				//!< @brief  Optional for application use. 
	int item3;				//!< @brief  Optional for application use. 
	char c_item1;			//!< @brief  Optional for application use. 
	char c_item2;			//!< @brief  Optional for application use. 
	char c_item3;			//!< @brief  Optional for application use. 
	char c_item4;			//!< @brief  Optional for application use. 
} GATCP_MSG_HDR;

/**
 * @brief PROXY_INFO
 */
typedef struct {
	char server[500]; //!< @brief server 
	int port;//!< @brief  port
	char userid[100];//!< @brief user id 
	char password[100];//!< @brief  password
	char useragent[100];//!< @brief user agent
} PROXY_INFO;

/**
 * @brief GATCP_SOCKET_MODE
 */
typedef enum {
	GATCP_MODE_BLOCKING = 0, //!< GATCP_MODE_BLOCKING 
	GATCP_MODE_NONBLOCK //!<  GATCP_MODE_NONBLOCK 
} GATCP_SOCKET_MODE;

/**
 * @brief GATCP_STATUS_T
 */
typedef enum {
	GATCP_STS_ERROR = 0, //!< GATCP_STS_ERROR 
	GATCP_STS_NO_DATA, //!< GATCP_STS_NO_DATA
	GATCP_STS_SUCCESS, //!< GATCP_STS_SUCCESS
	GATCP_STS_NEW_CONN //!< GATCP_STS_NEW_CONN 
} GATCP_STATUS_T;

/**
 * @struct gatcp_socket_info
 * @brief structure
 */
typedef struct gatcp_socket_info {
	int sockfd;//!< @brief socket file discriptor
	struct gatcp_socket_info *prev; //!< @brief previous
	struct gatcp_socket_info *next; //!< @brief next
} GATCP_SOCKET_INFO;//!< @brief typedef

/* In $lib/comm/gatcp.c */
/**
 * @brief tcp_init 
 * @param hostip
 * @param port
 * @param make_blocking
 * @param rcvbuf_size
 * @param sndbuf_size
 * @return
 */
int tcp_init (char *hostip, int port, int make_blocking, int rcvbuf_size,
	                int sndbuf_size);

/**
 * @brief tcp_init_conn
 * @param cinfo
 * @return
 */
int tcp_init_conn (GATCP_CONN_INFO *cinfo);

/**
 * @brief set socket mode
 * @param fd
 * @param mode
 * @return
 */
int set_socket_mode (int fd, GATCP_SOCKET_MODE mode);

/**
 * @brief tcp_read
 * @param fd
 * @param data
 * @param dsize
 * @return
 */
int tcp_read (int fd, void *data, int dsize);

/**
 * @brief tcp readline
 * @param fd
 * @param ptr
 * @param maxlen
 * @param seconds
 * @return
 */
int tcp_readline (int fd, char *ptr, int maxlen, int seconds);

/**
 * @brief tcp_write
 * @param fd
 * @param data
 * @param nbytes
 * @return
 */
int tcp_write (int fd, void *data, int nbytes);

/**
 * @brief tcp send message
 * @param cinfo
 * @param hdr
 * @param data
 * @param timeout
 * @return
 */
int tcp_send_msg (GATCP_CONN_INFO *cinfo, GATCP_MSG_HDR *hdr, void *data,
					GATIME *timeout);

/**
 * @brief try to reconnect
 * @param cinfo
 * @return
 */
int try_to_reconnect (GATCP_CONN_INFO *cinfo);

/**
 * @brief Caller specifies bufsize of 0 when they want routine to malloc space.
 * @param cinfo
 * @param hdr
 * @param data
 * @param bufsize
 * @param timeout
 * @return
 */
GATCP_STATUS_T tcp_recv_msg (GATCP_CONN_INFO *cinfo, GATCP_MSG_HDR *hdr,
					void **data, int *bufsize, GATIME *timeout);

/**
 * @brief Caller specifies bufsize of 0 when they want routine to malloc space.
 * @param cinfo
 * @param hdr
 * @param data
 * @param bufsize
 * @param timeout
 * @param extra_bytes;
 * @return
 */
GATCP_STATUS_T tcp_recv_msg2 (GATCP_CONN_INFO *cinfo, GATCP_MSG_HDR *hdr,
					void **data, int *bufsize, GATIME *timeout, int extra_bytes);

/**
 * @brief get local ip
 * @param host_ip
 * @param hlen
 * @param err_msg
 * @return
 */
int get_local_ip (char *host_ip, int hlen, char *err_msg);

/**
 * @brief host to ip
 * @param h_str
 * @param host_addr
 */
void host2ip (char *h_str, char *host_addr);

/**
 * @brief gatcp mod socket list
 * @param socklist
 * @param sockfd
 * @param del_socket
 * @return 
 */
int gatcp_mod_socket_list (GATCP_SOCKET_INFO **socklist, int sockfd,
			    int del_socket);

/**
 * @brief gatcp hdr checksum
 * @param hdr
 * @return
 */
uint32_t /*unsigned long*/ gatcp_hdr_checksum(GATCP_MSG_HDR *hdr);

/**
 * @brief Call before tcp init to set port to bind to during tcp_init as client 
 * @param port
 * @return
 */
int tcp_set_local_port (int port);

#ifdef	__cplusplus
}
#endif

#endif /* GATCP_H */
