/*
	tool for using xpat.  lets one diconnect a bit from the xpat 
	interface by passing a data structure and callbacks to this 
	code and you get the data you want.  
	
	i am told this is somewhat like a SAX model.

	build a table of xpatool_Tag structures.  Each item 'describes' a tag
	that you have interest in.  You provide a callback function that
	is called with the attributes of the tag and the data bound by the
	tag.  this callback is done when the end tag, </tag>, is seen.

	for collecting all of the data and tags bound by a specific tag, for
	example all of the story text, you set the flag 'xpaGatherAll', it
	is bit 0 so do something like 
	
	tagflags |= (1 << xpatoolGatherAll);
	or
	BIT_SET(tagflags ,(1 << xpatoolGatherAll));

	to set the flag in the tag table entry.  
	
	NOTE: do not set xpaGatherAll flags on two tags that are nested!
	you will break things badly!

    copyright 2005	Acquire Media Corp,
    
	20051209		glw		initial build
	20060209        glw     some name changes...

*/

#if !defined __xpatool_header__
#define __xpatool_header__

/* ///////////////////////////////////////////////////////////////////////// */
#ifdef	__cplusplus
extern "C" {
#endif

/* process bits for the tag, determines what we want to do with the tag */
typedef enum {
	xpatoolGatherAll = 0  /* collect all tags and data e.g. for story data*/

} xpatool_Flags;


/* sizes of tag attributes. */
#define xpatool_MAXATTSTAG (32)  /* maximum attributes per tag */
#define xpatool_MAXATTLENGTH (256)  /* maximum attribute length */

/* callback for a tag...  called when end of tag is seen and all of the
	data etc. has been collected and can be passed to the routine. 

	Notes:  
	-tag path looks like fred<tom<harry<james ...
	where fred is the root tag and successive tags separated by <
	-tagdata is the data collected between <tag> and </tag>
	-tagdatasize is the number of bytes at tagdata
	-attributes are a list of pointers to strings, a null ptr ends
	the list.  attributes are pairs, the tag attribute fred="joe"
	is returned as two succesive character strings "fred" then "joe"
*/
typedef void (*xpatool_TagCB)(const char *tagpath  /* tag path for this tag */
					,const char *tagdata  /* collected data of the tag */
					,int tagdatasize  /* size of the tag data */
					,char **attributes);  /* list of tag attributes */

/* tag information, build an entry for each tag that you want to deal with
	unless an enclosing tag's info has the 'GatherAll' process bit set, which
	causes us to collect the enclosed tags and data for later use! 
*/
typedef struct __xpatag_tag__ {
	/* name of the tag ! case sensitive ! and its length */
	const char *tagtext;
	const size_t tagtxtlen;
	/* numeric value for this tag */ 
	const int tagval;
	/* process bits */
	const unsigned long tagflags;
	/* processing call back function pointer; called when the closing 
		tag is seen.  passed all of the collected information for this
		tag etc...  see above for details... */
	const xpatool_TagCB tagcallback;

} xpatool_Tag;

/* tool structure passed to xpaProcess  */
typedef struct __xpatool_tag__ {
	size_t count;
	xpatool_Tag *tagitems;

} xpatool;

/* ////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////// */
/* xpaTool functions */


/* process... returns 0 on failure to process else 1 
	this function calls on the xpat parser to get at the xml file.  the tool
	table is used to collect data and call processing callback functions for
	the data as the file is parsed.

	Notes:
		-tool pointer to the table of tags in which you are interested
		-filename  a unique name for this parse.  xpaProcess can handle
		up to 5 simultaneous operations
		-buffer pointer to the data to process
		-bufferlen  amount of data to process
		-complete  if true everything is done in this buffer else you send
		another buffer in a subsequent call
*/
extern int xpatool_Process(xpatool *tool /* tool used to process the data */
						   ,const char *filename /* file name being processed */
						   ,unsigned char *buffer /* data to process */
						   ,int bufferlen  /* amount of data */
						   ,int complete); /* complete data: 1=>all 0=>more */

/* if xpaProcess returns 0 these are the errors from the xpat call... 
   this is good if the failure was from xpat not other conditions.

*/
extern void xpatool_LastErrors(int *numeric ,const char **text ,int *line_number);


#ifdef	__cplusplus
}
#endif
/* ///////////////////////////////////////////////////////////////////////// */

#endif /* defined __xpatool_header__ */

/*/////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////// */
/* end of file */









