#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <time.h>
#include <errno.h>

#include "apputil.h"
#include "wndpcomm.h"
#include "xmlparse/xmlnews.h"

static void usage (void);
static void setup_cfg (void);
static int process_message (WNDP_MESSAGE *msg);
static int process_rdf (WNDP_FILE_INFO *rdf_info);
static int process_xml (WNDP_FILE_INFO *xml_info);
static void print_viewer_time (void);
static void handle_error (void);

static void sig_pipe (int signum);
static void sig_exit (int signum);

/* To be read from configuration file. */
static char WNDP_Server[200] = "";
static int WNDP_Port = -1, PrintViewerTime;
static int ReadStory = FALSE;
WNDP_FLTR_MODE FilterMode = WNDP_FLTR_INCLUDE;
WNDP_FLTR_ITEM FilterItems[2];
FILE *out_fp;

int main (int argc, char *argv[])
{
	int i, make_daemon = TRUE;
	char filename[200];
	WNDP_MESSAGE msg;

	out_fp = stdout;

	/* Save process name and directory from which process was run so
	can use to build default paths of things like the configuration
	file. */

	app_setup (argv[0], AMXVIEWER_REGISTRY_CFG);

	PrintViewerTime = FALSE;
	filename[0] = 0;
	for (i = 1; i < argc; i++) {
		if (strcmp (argv[i], "-h") == 0) {
			usage ();
		}
		else if (strcmp (argv[i], "-d") == 0) {
			make_daemon = FALSE;
		}
		else if (strcmp (argv[i], "-t") == 0) {
			PrintViewerTime = TRUE;
		}
		else if (strcmp (argv[i], "-f") == 0) {
			/* Output filename -- defaults to stdout */
			if (++i == argc) {
				usage ();
			}
			strcpy (filename, argv[i]);
		}
		else if (strcmp (argv[i], "-i") == 0) {
			/* User specified WNDP server. */
			if (++i == argc) {
				usage ();
			}
			strcpy (WNDP_Server, argv[i]);
		}
		else if (strcmp (argv[i], "-p") == 0) {
			/* Provider for FLTR command */
			if (++i == argc) {
				usage ();
			}
			strcpy (FilterItems[0].provider, argv[i]);
		}
		else if (strcmp (argv[i], "-r") == 0) {
            ReadStory = TRUE;
        }
	}

	if (filename[0] != 0) {
		out_fp = fopen (filename, "w");
		if (out_fp == NULL) {
			plog (NL_CRIT, "Unable to open %s (%s).", filename,
					strerror (errno));
			sig_exit (APP_ERROR);
		}
	}


	/* Read the configuration file and setup all configuration parameters. */
	setup_cfg ();

	signal (SIGINT, sig_exit);
	signal (SIGTERM, sig_exit);
	signal (SIGBUS, sig_exit);
	signal (SIGSEGV, sig_exit);
	signal (SIGPIPE, sig_pipe);

	/* Turn process into a daemon. */
	if (make_daemon && out_fp != stdout) {
		if (! create_daemon ()) {
			sig_exit (APP_ERROR);
		}
	}
	plog (NL_STATUS, "PID: %ld started.", getpid ());

	while (TRUE) {

		plog (NL_STATUS, "Opening TCP connection to WNDP Server.");
		if (! wndp_connect (WNDP_Server, WNDP_Port, 4)) {
			handle_error ();
			continue;
		}

		if (! wndp_login ("USER1", "")) {
			handle_error ();
			continue;
		}

		/* Turn on all data classes. */
		if (! wndp_set_classes ("ALL")) {
			handle_error ();
			continue;
		}

		if (! wndp_set_fltr (FilterMode, FilterItems)) {
			handle_error ();
			continue;
		}

		/* Get each message from server. Call with timeout of 0 seconds
		so we block until server actually has data. */

		while (wndp_get_next_message (&msg, 0)) {

			if (! process_message (&msg)) {
				break;
			}
		}

		/* Error communicating with server. Disconnect and loop back
		to reconnect. */
		wndp_disconnect ();
		plog (NL_STATUS, "Force new connection to WNDP server.");
	}
}

static void usage (void)
{
	printf ("\n    %s [-h|-d|-i <server>|-f <filename>|-p <provider>|-t|-r]\n\n"

		"\t-h               - Prints this message.\n"
		"\t-d               - Runs process in the forground for debugging.\n"
		"\t-i <server>      - Specifies WNDP server to connect to.\n"
		"\t                   Default is the current server.\n"
		"\t-f <filename>    - Specifies filename to which output should be\n"
		"\t                   written.\n"
		"\t-p <provider>    - Specifies name of provider whose headlines\n"
		"\t                   should be displayed by this process.\n"
		"\t-t               - Specifies that viewer should print the\n"
		"\t                   receive time in milliseconds for each message.\n"
		"\t-r               - Specifies that viewer should read content\n"

		"\n", ProcessName);

	exit (1);
}

static void setup_cfg (void)
{
	FILE *fp;
	char cfg_file[200], line[200], *value;

	sprintf (cfg_file, "%s/cfg/%s.cfg", ProductDir, ProcessName);
	if (! file_exists_time (cfg_file, NULL)) {
		/* No cfg file. Use all default values. */
		return;
	}
	fp = fopen (cfg_file, "r");
	if (fp == NULL) {
		printf ("\n    Unable to open %s (%s)\n", cfg_file, strerror (errno));
		exit (APP_ERROR);
	}

	while (fgets (line, sizeof (line), fp) != NULL) {
		rtrim (line);
		if (line[0] == '#' || line[0] == 0) {
			/* Skip comments and blank lines. */
			continue;
		}
		value = strchr (line, ':');
		if (value == NULL) {
			continue;
		}
		*value = 0;

		/* Trim leading white space from value. */
		value = ptr_ltrim (value + 1);

		if (strcmp (line, "WNDP_Server") == 0) {
			strcpy (WNDP_Server, value);
		}
		else if (strcmp (line, "WNDP_Port") == 0) {
			WNDP_Port = atoi (value);
		}
	}
	fclose (fp);
}

static int process_message (WNDP_MESSAGE *msg)
{
	int i;
	char *s, buf[200];

	/* Look for the rdf file in the file list and process it when
	it is found. */

	for (i = 0; i < msg->nfiles; i++) {
		s = msg->files[i].name + strlen (msg->files[i].name) - 4;
		if (strcmp (s, ".rdf") == 0) {
			return process_rdf (&msg->files[i]);
		}

		if (strcmp (msg->data_class, "XMLNEWS-STORY") != 0) {
		    s = msg->files[i].name;
		    sprintf (buf, "%2.2s:%2.2s  %-8s  File: %s (%d bytes)\n",
					s+8, s+10, msg->data_class, s, msg->files[i].size);
		    fputs (buf, out_fp);
			if (PrintViewerTime) {
				print_viewer_time ();
			}
		    fflush (out_fp);

            if (ReadStory) {
                process_xml (&msg->files[i]);
            }
		}
	}

	/* No rdf file in this message. Return TRUE so caller goes for next
	set of files. */
	return TRUE;
}

/* process_rdf returns FALSE only when an error occurs during
communication with the WNDP Server. */

static int process_rdf (WNDP_FILE_INFO *rdf_info)
{
	char line[300], buf[30];
	XMLNEWS_META *xmlnews_meta;


	if (! wndp_get_file (rdf_info)) {
		handle_error ();
		return FALSE;
	}

	/* Parse the xmlnews_meta data to extract the headline and the
	receive time.  Note that the definition of XMLNEWS_META may change
	in the future. */

	xmlnews_meta = xmlnews_parse_meta (rdf_info->name, rdf_info->data,
										rdf_info->size, TRUE);
	if (xmlnews_meta == NULL || xmlnews_meta->title == NULL) {
		/* Could not parse rdf file via expat, but return TRUE so
		caller goes on to get next file. */
		return TRUE;
	}

	/* Clear line to facilitate formatting output in fixed fields. */
	memset (line, ' ', sizeof (line));

	strncpy (line, xmlnews_meta->publicationTime + 11, 5);

	/* Put 8 character provider name from filename into output line. */
	strncpy (line + 5 + 2, rdf_info->name + 12, 8);

	/* Add headline to end. */
	strncpy (line + 5 + 2 + 8 + 2, xmlnews_meta->title, 60);

	/* Trim trailing blanks from line and add newsline. */
	line[sizeof (line) - 1] = 0;
	rtrim (line);
	strcat (line, "\n");

	sprintf (buf, "       %4.4s%2.2s%2.2s  ",
	    xmlnews_meta->publicationTime,
	    xmlnews_meta->publicationTime + 5,
	    xmlnews_meta->publicationTime + 8);
	strcat (line, buf);
	strcat (line, "(receivedTime: ");
	strcat (line, xmlnews_meta->receivedTime);
	strcat (line, ")\n");

	fputs (line, out_fp);

	if (PrintViewerTime) {
		print_viewer_time ();
	}

	fflush (out_fp);

	/* Free allocated buffer space. */
	free (rdf_info->data);
	rdf_info->data = NULL;

	return TRUE;
}

static int process_xml (WNDP_FILE_INFO *xml_info)
{

	if (! wndp_get_file (xml_info)) {
		handle_error ();
		return FALSE;
	}

    printf ("%s\n", xml_info->data);

	/* Free allocated buffer space. */
	free (xml_info->data);
	xml_info->data = NULL;

	return TRUE;
}

static void print_viewer_time (void)
{
	long tod;
	char time_buf[100];
	struct timeval now;

	gettimeofday (&now, NULL);
	tod = now.tv_sec;
	strftime (time_buf, sizeof (time_buf), "%T", localtime(&tod));
	sprintf (time_buf + strlen (time_buf), ".%03ld ", now.tv_usec/1000);
	fprintf (out_fp, "       Viewer Time: %s\n", time_buf);
}

static void handle_error (void)
{
	char *s;

	s = wndp_get_err_str ();
	if (*s != 0) {
		plog (NL_CRIT, "%s", s);
		if (out_fp == stdout) {
			printf ("ERROR: %s\n", s);
			fflush (stdout);
		}
		sig_exit (APP_ERROR);
	}
}

static void sig_pipe (int signum)
{
	/* SIGPIPE occurs when write to the socket after the server has
	exited. Default action would be for process to exit. */
	wndp_disconnect ();
	plog (NL_STATUS, "Force new connection to WNDP server.");
}

static void sig_exit (int signum)
{
	wndp_disconnect ();
	app_exit (signum);
}
