#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <errno.h>
#include <time.h>
#include <wait.h>
#include <stdarg.h>
#ifdef _REENTRANT
	#include <pthread.h>
#endif

#include "apputil.h"

static int plog_init (struct tm *tp_today);

char ProcessName[40] = "", ProductDir[200] = "", TmpDir[200] = "";

static char LogName[40] = "", LogDir[200] = "";
static int LogLvl = NL_STATUS;
static FILE *plog_fp = NULL;
static int nchildren = 0;


void rtrim (char *instr)
{
	int i;

	/* Strip all white space (blanks, tabs, newlines, form feed) from end. */
	i = strlen (instr);
	while (i > 0 && isspace ((int) instr[i-1])) {
		instr[--i] = '\0';
	}
}

char *ptr_ltrim (char *instr)
{
	char *s;

	s = instr;
	while (isspace ((int) *s)) {
		s++;
	}
	return s;
}

int plog (int level, ...)
{
	va_list	args;
	int fd;
	char *fmt, fmt_str[1000], levstr[20], timestr[20];
	long tod;
	struct tm tp;

	static char plog_lockfile[200] = "";
	static int weekday = -1;

	#ifdef _REENTRANT
		int status;
		static pthread_mutex_t mlock = PTHREAD_MUTEX_INITIALIZER;
	#endif


	/* Validate passed in level. */
	if (level > NL_WARN && level > LogLvl) {
	    /* Caller passed NL_STATUS, NL_INFO, ... or lower level */
	    return TRUE;
	}


	#ifdef _REENTRANT
		/* Get a mutex lock so that the various threads within this
		process do not collide with each other. We will also need to
		get a filelock so that this process and other processes logging
		to this same log file do not collide. */

		if ((status = pthread_mutex_lock (&mlock)) != 0) {
			/* Could not get mutex. CANNOT use plog within plog --
			creates infinite recursion. Return 'cause there is nothing
			else we can do. */
			printf ("Unable to get mutex lock (%s)", strerror (status));
			return TRUE;
		}
	#endif

	/* Get an advisory lock on the plog lock file so that the various
	processes writing to this log file do not collide with each other.
	Last argument of TRUE means that function will block until the file
	is available to this process. */

	if (plog_lockfile[0] == 0) {
		sprintf (plog_lockfile, "%s/pLog.lock", TmpDir);
	}
	if (! filelock_getlock (plog_lockfile, &fd, TRUE)) {
		/* CANNOT use plog within plog -- creates infinite recursion.
		Return 'cause there is nothing else we can do. */
		printf ("Unable to get lock file %s (%s)", plog_lockfile,
				strerror (errno));
		#ifdef _REENTRANT
			/* Release the mutex. */
			pthread_mutex_unlock (&mlock);
		#endif
	    return TRUE;
	}


	va_start (args, level);
	fmt = va_arg (args, char *);

	/* Write process name to output string. */
	#ifdef _REENTRANT
		sprintf (fmt_str, "%s%lu: ", LogName, pthread_self());
	#else
		sprintf (fmt_str, "%s: ", LogName);
	#endif

	/* Print the message passed in. */
	if (fmt != (char *) NULL) {
		vsprintf (fmt_str + strlen (fmt_str), fmt, args);
	}
    va_end (args);

	switch (level) {
	case NL_CRIT:
	    strcpy (levstr, "CRITICAL: ");
		break;
	case NL_WARN:
	    strcpy (levstr, "WARNING:  ");
		break;
	case NL_STATUS:
	    strcpy (levstr, "STATUS:   ");
		break;
	case NL_INFO:
	    strcpy (levstr, "INFO:     ");
		break;
	case NL_DEBUG:
	    strcpy (levstr, "DEBUG:    ");
		break;
	default:
	    strcpy (levstr, "****:     ");
		break;
    }

    tod = time ((time_t *) 0);
	localtime_r (&tod, &tp);
	strftime (timestr, sizeof (timestr) - 1, "%H:%M:%S", &tp);

	if (tp.tm_wday != weekday || plog_fp == NULL) {
		/* It's a different day of the week. Time to open a new log file. */
		plog_init (&tp);
		weekday = tp.tm_wday;
	}

	/* Make sure that message (fmt_str) does not have any trailing
	white space or new lines. */
	rtrim (fmt_str);

	/* Write message to log file. Seek to EOF since other processes
	may have written messages since the last time this process did. */
	fseek (plog_fp, 0, SEEK_END);
	fprintf (plog_fp, "%s %s %s\n", timestr, levstr, fmt_str);
	fflush (plog_fp);

	/* Release the file lock. */
	filelock_unlock (fd);

	#ifdef _REENTRANT
		/* Release the mutex. */
		pthread_mutex_unlock (&mlock);
	#endif

	return TRUE;
}

static int plog_init (struct tm *tp_today)
{
	FILE *fp;
	char logfile[200], mode[5] = "w";
	long mtime;
	struct tm tx;
	static char *days[] = {"sun", "mon", "tue", "wed", "thu", "fri", "sat"};


	/* Open correct log file making sure not to overwrite a file of
	the same name if it is from today. */
	sprintf (logfile, "%s/%s", LogDir, days[tp_today->tm_wday]);
	if (file_exists_time (logfile, &mtime)) {
		localtime_r (&mtime, &tx);
		if (tx.tm_yday == tp_today->tm_yday) {
			/* File is from today. Open it in append mode. */
			strcpy (mode, "a");
		}
	}
	fp = fopen (logfile, mode);
	if (fp == NULL) {
		/* CANNOT use plog within plog -- creates infinite recursion. */
		printf ("Unable to open %s (%s)\n", logfile, strerror (errno));
		return FALSE;
	}
	if (plog_fp != NULL) {
		fclose (plog_fp);
	}
	plog_fp = fp;

	if (*mode == 'w') {
		/* New file. Write today's date at the top. Use logfile as a
		temp var. */
		strftime (logfile, sizeof (logfile) - 1, "Messages for %D :\n\n",
					tp_today);
		fprintf (plog_fp, "%s", logfile);
		fflush (plog_fp);
	}
	return TRUE;
}

void change_logname (char *new_logname)
{
	strncpy (LogName, new_logname, sizeof (LogName) - 1);
	LogName[sizeof (LogName) - 1] = 0;
}

void change_loglvl (int new_lvl)
{
	LogLvl = new_lvl;
}

int create_daemon (void)
{
	int i, fd;
    long pid;
	char fdname[200];

    pid = fork ();
	if (pid < 0) {
		plog (NL_CRIT, "Failed to fork (%s)", strerror (errno));
		return FALSE;
	}
	if (pid != 0) {
		/* Parent process, just exit. */
		exit (0);
	}

    /* Keep stdout and stderr from being used. */
    close (1); 
    close (2); 

	/* Reuse file descriptors 0, 1, 2 so that renegade printf's
	in code do not cause arbitrary file corruptions. */

	for (i = 1; i < 3; i++) {

		sprintf (fdname, "%s/%s.fd%d", TmpDir, ProcessName, i);

		/* Remove in case this exists by another user from same group. */
		unlink (fdname);

		if ((fd = open (fdname, O_WRONLY | O_CREAT | O_TRUNC)) < 0) {
			plog (NL_CRIT, "Could not open %s", fdname);
			return FALSE;
		}
		if (fd != i) {
			plog (NL_CRIT, "fd%d is equal to %d, not %d.", i, fd, i);
			return FALSE;
		}

		/* Set file permissions to user and group read, write. */
		if (chmod (fdname, S_IRUSR | S_IRGRP | S_IWUSR | S_IWGRP) == -1) {
			plog (NL_CRIT, "Failed chmod (%s)", strerror (errno));
			return FALSE;
		}
	}

    setsid (); 

    /* Ignore process group leader death. */
    signal (SIGUSR2, SIG_IGN);

	return TRUE;
}

int app_mkpath (char *path)
{
	char *s, *tmp;

	/* Make sure that all directories in the path exist. Skip a leading
	slash since we cannot create the root partition. This functions
	works like the cshell command 'mkdir -p' */

	tmp = path + 1;
	while ((s = strchr (tmp, '/')) != NULL) {
		*s = 0;
		if (! app_mkdir (path)) {
			return FALSE;
		}
		*s = '/';
		tmp = s + 1;
	}
	if (! app_mkdir (path)) {
		return FALSE;
	}
	return TRUE;
}

int app_mkdir (char *dir_name)
{

	if (dir_exists (dir_name)) {
		/* Directory already exists. */
		return TRUE;
	}

	/* Directory does not exist yet. Create it (mode 1FF = 777). */
	if (mkdir (dir_name, 0x1FF) != -1) {
		/* Success.  Note, umase will possible effect the directories
		permissions, so explicitly set them here. */
		chmod (dir_name, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
		return TRUE;
	}
	plog (NL_WARN, "Cannot create directory %s (%s)", dir_name,
			strerror (errno));
	return FALSE;
}

int file_exists_time (char *file, long *mtime)
{
    struct stat buf;
     
    if (0 == stat (file, &buf)) {
		if (mtime != NULL) {
			*mtime = buf.st_mtime;
		}
		return S_ISREG(buf.st_mode);
    }
    return FALSE;
}

int dir_exists (char *dir)
{
    struct stat buf;
     
    if (0 == stat (dir, &buf)) {
		return S_ISDIR (buf.st_mode);
    }
    return FALSE;
}

long get_file_size (char *file)
{
    struct stat buf;
     
    if (0 == stat (file, &buf)) {
		return buf.st_size;
    }
    return -1;
}

void app_setup (char *argv0, char *registry_cfg_file)
{
	char *s;
	READ_CFG_STS sts;

	sts = read_cfg_file (ProductDir, sizeof (ProductDir),
			"MNTPNT:", registry_cfg_file);
	if (sts != READ_CFG_FOUND) {
		printf ("Failed processing file %s\n", registry_cfg_file);
		app_exit (APP_ERROR);
	}
	/* All products reside in "programs" below MNTPNT: */
	strcat (ProductDir, "/programs");

	s = strrchr (argv0, '/');
	if (s == NULL) {
		strcpy (ProcessName, argv0);
	}
	else {
		strcpy (ProcessName, s + 1);
	}

	/* Use this info to establish values for LogName, LogDir, TmpDir.
	Also make sure the directories exist. */
	strcpy (LogName, ProcessName);

	sprintf (LogDir, "%s/logs", ProductDir);
	app_mkpath (LogDir);

	strcpy (TmpDir, ProductDir);
	s = strrchr (TmpDir, '/');
	if (s == NULL) {
		printf ("Unable to perform directory setup for TmpDir.\n");
		app_exit (APP_ERROR);
	}
	strcpy (s + 1, "tmp");
	app_mkpath (TmpDir);
}

void app_exit (int signum)
{   
	int plogsts = NL_STATUS;

	if (signum != SIGBUS && signum != SIGSEGV) {
		if (signum == APP_ERROR) {
		    plogsts = NL_CRIT;
		}
		plog (plogsts, "Process exiting with status %d.", signum);
		exit (signum);
	}

	plog (NL_CRIT, "Process exiting with status %d.", signum);

	/* Only get here for SIGBUS and SIGSEGV. */

	/* Change directory so that all core files can be found. */
	chdir (TmpDir);
	abort ();
}

void increment_child_count (void)
{
	nchildren++;
}

int ready_for_child (int max_children)
{
	long tod;
	struct tm *tp;
	static int hour = -1;

	/* For some reason (yet to be uncovered), nchildren is sometimes
	larger than it should be. Reset it to zero once every hour. This
	will result in potentially greater than max_children child processes
	for a very short while, but the number of processes will still be
	bounded. */

    tod = time ((time_t *) 0);
	tp = localtime (&tod);
	if (tp->tm_hour != hour) {
		nchildren = 0;
		hour = tp->tm_hour;
	}

	return (nchildren < max_children);
}

void app_sigchld (int signum)
{
	int status, exit_val;
	long pid;

	while ((pid = waitpid ((pid_t) -1, &status, WNOHANG)) > 0) {

		if (nchildren > 0) {
			nchildren--;
		}
		else {
			plog (NL_DEBUG, "Unexpected non-positive nchildren = %d",
					nchildren);
		}

		if (WIFEXITED (status)) {
		    
			/* Child process exited via call to exit () */

			exit_val = WEXITSTATUS (status);
			if (exit_val != 0 && exit_val != SIGTERM && exit_val != SIGPIPE) {
				plog (NL_WARN, "Child process terminated with exit code %d.",
						exit_val);
			}
		}
		else {
			plog (NL_WARN, "Child process terminated unexpectedly.");
		}
	}
}


/* Reliable version of signal(), using POSIX sigaction. Taken from
Advanced Programning in the UNIX Environment by W. Richard Stevens.
If this version of signal() is not used, each signal handler must
re-establish itself as required. We STRONGLY advise the use of this
version of signal(). */

SigFunc *signal (int signo, SigFunc *func)
{
    struct sigaction act, oact;

	act.sa_handler = func;
	sigemptyset (&act.sa_mask);
	act.sa_flags = 0;

	if (sigaction (signo, &act, &oact)< 0) {
		return (SIG_ERR);
	}
	else {
		return (oact.sa_handler);
	}
}

READ_CFG_STS read_cfg_file (char *buf, int bufsz, char *label, char *filename)
{
	FILE *fp;
	char tmp[5000], *s, *p;
	int line_len, found = FALSE;


	if (! file_exists_time (filename, NULL)) {
		return READ_CFG_NOFILE;
	}

	if ((fp = fopen (filename, "r")) == NULL) {
		plog (NL_WARN, "Failed to open %s (%s)", filename, strerror (errno));
		return READ_CFG_FAILURE;
	}

	/* Look for specified line label in the configuration file. */
	line_len = strlen (label);
	while (found == FALSE) {

		if (fgets (tmp, sizeof (tmp), fp) == NULL) {
			break;
		}

		/* Remove trailing newline. */
		s = strchr (tmp, '\n');
		if (s != NULL) {
			*s = 0;
		}

		/* Remove leading white space. */
		s = tmp;
		ptr_ltrim (s);

		/* Look for end of line label. */
		p = strpbrk (s, " \t");

		if (p != NULL) {
			*p = 0;
			if (strcmp (label, s) == 0) {
				/* Found match. */
				p++;

				/* Trim answer and remove any comment that may be
				appended on the same line starting with '#' */

				ptr_ltrim (p);
				if ((s = strrchr (p, '#')) != NULL) {
					*s = 0;
				}
				rtrim (p);

				if (strlen (p) >= (size_t) bufsz) {
					plog (NL_WARN, "Data for %s too long for buffer",
							label);
					break;
				}
				strcpy (buf, p);
				found = TRUE;
			}
		}
	}

	fclose (fp);

	return (found ? READ_CFG_FOUND : READ_CFG_NOTFOUND);
}


/* statics for ignore_signals and restore_signals. Note ignore_signals
and restore_signals are NOT written for use in threaded applications. */

static struct sigaction ignore, saveintr, saveterm;
static int save_sigintr = 0, save_sigterm = 0;
static int ignore_level = 0;

static void tmp_sighandler (int signo)
{
	switch (signo) {
		case SIGINT:
			save_sigintr = signo;
			break;
		case SIGTERM:
			save_sigterm = signo;
			break;
		default:
			break;
	}
}

int ignore_signals (void)
{
	/* Have to keep track of how many times ignore_signal has been
	called since we do not want to overwrite the saved signal handlers
	that are stored during the first call. */
	if (ignore_level++ != 0) {
		/* We are already ignoring signals. Don't do anything. */
		return TRUE;
	}

	/* Ignore SIGINT, SIGTERM */

	save_sigintr = save_sigterm = 0;

	ignore.sa_handler = tmp_sighandler;
	sigemptyset (&ignore.sa_mask);
	ignore.sa_flags = 0;

	if (sigaction (SIGINT, &ignore, &saveintr) < 0) {
		plog (NL_WARN, "sigaction - ignore SIGINT (%s)", strerror (errno));
		return FALSE;
	}
	if (sigaction (SIGTERM, &ignore, &saveterm) < 0) {
		plog (NL_WARN, "sigaction - ignore SIGTERM (%s)", strerror (errno));
		return FALSE;
	}
	return TRUE;
}

int restore_signals (void)
{
	if (ignore_level == 0) {
		/* ignore_signals was never called. Nothing to do. */
		return TRUE;
	}
	if (--ignore_level > 0) {
		/* Haven't gotten back to the first ignore_signals level.
		Don't do anything yet. */
		return TRUE;
	}

	/* Restore previous signal action for SIGINT, SIGTERM. */

	if (sigaction (SIGINT, &saveintr, NULL) < 0) {
		plog (NL_WARN, "sigaction - restore SIGINT (%s)", strerror (errno));
		return FALSE;
	}
	if (sigaction (SIGTERM, &saveterm, NULL) < 0) {
		plog (NL_WARN, "sigaction - restore SIGTERM (%s)", strerror (errno));
		return FALSE;
	}

	/* Process SIGINT or SIGTERM if it was received while we were
	ignoring those signals. */
	if (save_sigintr == SIGINT) {
		saveintr.sa_handler (save_sigintr);
	}
	if (save_sigterm == SIGTERM) {
		saveterm.sa_handler (save_sigterm);
	}

	return TRUE;
}
